// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "algorithm.h"

namespace spex {

[[nodiscard]] auto binaryFindIdx(const std::vector<double>& list,
                                 const double x) -> int
{
    int i_begin {};
    int i_end { static_cast<int>(list.size() - 1) };
    int i_mid {};
    if (x <= list.front() || x >= list.back()) {
        return def::idx;
    } else if (x == list.front()) {
        return i_begin;
    } else if (x == list.back()) {
        return i_end;
    }
    while (true) {
        i_mid = (i_begin + i_end) / 2;
        if (x < list[i_mid]) {
            i_end = i_mid - 1;
        } else if (x < list[i_mid + 1]) {
            return i_mid;
        } else {
            i_begin = i_mid + 1;
        }
    }
}

auto updateMaskLt(const std::vector<double>& data,
                  const double threshold,
                  const std::string& info_str,
                  std::vector<bool>& mask) -> void
{
    updateMask(data, threshold, info_str, mask);
}

auto updateMaskGt(const std::vector<double>& data,
                  const double threshold,
                  const std::string& info_str,
                  std::vector<bool>& mask) -> void
{
    updateMask<std::greater<double>>(data, threshold, info_str, mask);
}

auto normalizeVector(std::array<double, dims::vec>& v) -> void
{
    const auto norm_inv { 1 / (std::max(1e-100, vectorNorm(v))) };
    for (auto& val : v) {
        val *= norm_inv;
    }
}

[[nodiscard]] auto dotProduct(const std::vector<double>::iterator& a,
                              const std::vector<double>::iterator& b) -> double
{
    return std::inner_product(a, a + dims::vec, b, 0.0);
}

[[nodiscard]] auto dotProduct(const std::array<double, dims::vec>& a,
                              const std::array<double, dims::vec>& b) -> double
{
    return std::inner_product(
      a.cbegin(), a.cbegin() + dims::vec, b.cbegin(), 0.0);
}

auto fillHoles(const std::vector<bool>& pixel_mask, std::vector<double>& image)
  -> void
{
    for (int i {}; i < static_cast<int>(image.size()); ++i) {
        if (pixel_mask[i]) {
            // Unless we are at either end of the image array, take
            // the average of the neighboring values.
            if (i == 0) {
                image.front() = image[1];
            } else if (i == static_cast<int>(image.size() - 1)) {
                image.back() = image[static_cast<int>(image.size()) - 2];
            } else {
                image[i] = (image[i - 1] + image[i + 1]) / 2;
            }
            // If there are two or more bad pixels side by side we
            // could still end up with a bad value after
            // averaging. Then we just set it to zero. To test it we
            // use a large value that is close to the NetCDF float
            // fill value.
            constexpr double bad_value { 1e30 };
            if (std::isnan(image[i]) || std::abs(image[i]) > bad_value) {
                image[i] = 0.0;
            }
        }
    }
}

} // namespace spex
