// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "bspline.h"

namespace spex {

BSpline::BSpline(const int order, const std::vector<double>& knots)
  : order { order }
{
    this->knots.resize(knots.size() + 2 * order);
    // Copy the B-spline knots and pad them from both ends by n
    // elements where n is the spline order.
    for (int i {}; i < static_cast<int>(knots.size()); ++i) {
        this->knots[i + order] = knots[i];
    }
    for (int i {}; i < order; ++i) {
        this->knots[i] = knots.front();
        this->knots[this->knots.size() - 1 - i] = knots.back();
    }
    control_points_tmp.resize(this->knots.size(), 0.0);
}

[[nodiscard]] auto BSpline::nStates() const -> int
{
    return static_cast<int>(knots.size()) - order - 1;
}

auto BSpline::findInterval(const double x) const -> int
{
    switch (knot_spacing) {
    case KnotSpacing::equal:
    case KnotSpacing::irregular:
    default:
        return binaryFindIdx(knots, x);
    }
}

auto BSpline::deBoor(const std::vector<double>& control_points,
                     const double x,
                     const int i_shift) const -> double
{
    const int p { findInterval(x) };
    if (p == def::idx) {
        return 0.0;
    }
    const auto control_points_begin { control_points.cbegin()
                                      + i_shift * nStates() };
    std::vector<double> d_cur { control_points_begin + p - order,
                                control_points_begin + p + 1 };
    for (int r { 1 }; r <= order; ++r) {
        for (int j { order }; j >= r; --j) {
            const double alpha { (x - knots[j + p - order])
                                 / (knots[j + 1 + p - r]
                                    - knots[j + p - order]) };
            d_cur[j] = (1 - alpha) * d_cur[j - 1] + alpha * d_cur[j];
        }
    }
    return d_cur[order];
}

auto BSpline::evalBasis(const double x,
                        const int i_row,
                        PackedMatrix& packed_matrix) -> void
{
    const int i_x { findInterval(x) };
    if (i_x == def::idx) {
        return;
    }
    std::vector<double> control_points(knots.size(), 0.0);
    std::vector<double> values(order + 1);
    for (int k {}; k <= order; ++k) {
        control_points[i_x - k] = 1.0;
        values[order - k] = deBoor(control_points, x);
        control_points[i_x - k] = 0.0;
    }
    packed_matrix.row_sizes[i_row] = order + 1;
    packed_matrix.global_col_indices[i_row] = i_x - order;
    packed_matrix.row_locations[i_row] =
      static_cast<int>(packed_matrix.data.size());
    for (auto& x : values) {
        packed_matrix.data.push_back(x);
    }
}

auto BSpline::evalBasis(const std::vector<double>& x_data,
                        std::vector<double>& B) -> void
{
    const int n_states { nStates() };
    B.resize(x_data.size() * n_states);
    for (int i_point {}; i_point < static_cast<int>(x_data.size()); ++i_point) {
        const int i_x { findInterval(x_data[i_point]) };
        if (i_x == def::idx) {
            return;
        }
        for (int k {}; k <= order; ++k) {
            control_points_tmp[i_x - k] = 1.0;
            B[i_point * n_states + i_x - k] =
              deBoor(control_points_tmp, x_data[i_point]);
            control_points_tmp[i_x - k] = 0.0;
        }
    }
}

auto BSpline::genDerivControlPoints(
  const std::vector<double>& control_points,
  std::vector<double>& control_points_deriv) const -> void
{
    // The state vector (number of control points) of the
    // derivative B-spline is one less.
    const int deriv_n_states { nStates() - 1 };
    // If control_points contains several sets of control points,
    // iterate over all of them.
    for (int i_set {};
         i_set < static_cast<int>(control_points.size() / nStates());
         ++i_set) {
        const int shift { i_set * nStates() };
        const int shift_deriv { i_set * deriv_n_states };
        for (int i {}; i < deriv_n_states; ++i) {
            control_points_deriv[shift_deriv + i] =
              order
              * (control_points[shift + i + 1] - control_points[shift + i])
              / (knots[order + i + 1] - knots[i + 1]);
        }
    }
}

} // namespace spex
