// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "linalg.h"

#include <cmath>
#include <numeric>

namespace spex {

auto dppsv(const std::vector<double>& data,
           const std::vector<double>& std_devs,
           std::vector<double>& B,
           std::vector<double>& state_vector) -> void
{
    // Number of state parameters
    const int n_states { static_cast<int>(B.size() / data.size()) };
    // Scale each row of data by noise (standard deviations)
    for (int i {}; i < static_cast<int>(data.size()); ++i) {
        for (int j {}; j < n_states; ++j) {
            B[i * n_states + j] /= std_devs[i];
        }
    }
    // B^T x B
    std::vector<double> BTB(n_states * n_states, 0.0);
    for (int i {}; i < n_states; ++i) {
        for (int j {}; j < n_states; ++j) {
            for (int k {}; k < static_cast<int>(data.size()); ++k) {
                BTB[i * n_states + j] +=
                  B[k * n_states + i] * B[k * n_states + j];
            }
        }
    }
    // B^T x Y where Y is the data vector
    std::vector<double> BTY(static_cast<int>(data.size()), 0.0);
    for (int i {}; i < n_states; ++i) {
        for (int k {}; k < static_cast<int>(data.size()); ++k) {
            BTY[i] += B[k * n_states + i] * data[k] / std_devs[k];
        }
    }
    // Upper triangle of B^T x B
    std::vector<double> BTB_tri(n_states * (n_states + 1) / 2);
    for (int j {}; j < n_states; ++j) {
        for (int i {}; i <= j; ++i) {
            BTB_tri[i + (j + 1) * j / 2] = BTB[i * n_states + j];
        }
    }
    // Covariance matrix of linear inversion initialized to unity
    std::vector<double> Sx(n_states * n_states, 0.0);
    for (int i {}; i < n_states; ++i) {
        Sx[i * n_states + i] = 1.0;
    }
    constexpr char uplo { 'u' };
    int info {};
    dppsv_(
      &uplo, &n_states, &n_states, BTB_tri.data(), Sx.data(), &n_states, &info);
    state_vector.assign(n_states, 0.0);
    for (int i {}; i < n_states; ++i) {
        for (int k {}; k < n_states; ++k) {
            state_vector[i] += Sx[i * n_states + k] * BTY[k];
        }
    }
}

auto dppsv(const std::vector<double>& data,
           const std::vector<double>& std_devs,
           PackedMatrix& B,
           std::vector<double>& state_vector) -> void
{
    // Scale each row of data by noise (standard deviations)
    for (int i {}; i < B.nRows(); ++i) {
        for (int j {}; j < B.nCols(i); ++j) {
            B(i, j) /= std_devs[i];
        }
    }
    // B^T x B
    PackedMatrix B_T { B.nCols(), B.nRows() };
    B_T.transpose(B);
    std::vector<double> BTB(B.nCols() * B.nCols(), 0.0);
    B_T.multRightTransposed(B_T, BTB);
    // B^T x Y where Y is the data vector
    std::vector<double> BTY(B_T.nRows(), 0.0);
    for (int i {}; i < B_T.nRows(); ++i) {
        for (int j {}; j < B_T.nCols(i); ++j) {
            const int col_idx { B_T.globalCol(i, j) };
            BTY[i] += B_T(i, j) * data.at(col_idx) / std_devs.at(col_idx);
        }
    }
    // Number of state parameters
    const int n_states { B.nCols() };
    // Upper triangle of B^T x B
    std::vector<double> BTB_tri(n_states * (n_states + 1) / 2);
    for (int j {}; j < n_states; ++j) {
        for (int i {}; i <= j; ++i) {
            BTB_tri[i + (j + 1) * j / 2] = BTB[i * n_states + j];
        }
    }
    // Covariance matrix of linear inversion initialized to unity
    std::vector<double> Sx(n_states * n_states, 0.0);
    for (int i {}; i < n_states; ++i) {
        Sx[i * n_states + i] = 1.0;
    }
    constexpr char uplo { 'u' };
    int info {};
    dppsv_(
      &uplo, &n_states, &n_states, BTB_tri.data(), Sx.data(), &n_states, &info);
    state_vector.assign(n_states, 0.0);
    for (int i {}; i < n_states; ++i) {
        for (int k {}; k < n_states; ++k) {
            state_vector[i] += Sx[i * n_states + k] * BTY[k];
        }
    }
}

auto simpleLinearRegression(const std::vector<double>& x_data,
                            const std::vector<double>& y_data,
                            double& a,
                            double& b) -> void
{
    const int n_data { static_cast<int>(x_data.size()) };
    const double x_mean { std::accumulate(x_data.cbegin(), x_data.cend(), 0.0)
                          / n_data };
    const double y_mean { std::accumulate(y_data.cbegin(), y_data.cend(), 0.0)
                          / n_data };
    a = 0.0;
    double denominator {};
    for (int i {}; i < n_data; ++i) {
        const double x_diff { (x_data[i] - x_mean) };
        a += x_diff * (y_data[i] - y_mean);
        denominator += x_diff * x_diff;
    }
    a /= denominator;
    b = y_mean - a * x_mean;
}

auto simpleLinearRegression(const std::vector<double>& std_devs,
                            const std::vector<double>& x_data,
                            const std::vector<double>& y_data,
                            double& a,
                            double& b) -> void
{
    const int n_data { static_cast<int>(x_data.size()) };
    double w_sum {};
    double x_mean {};
    double y_mean {};
    for (int i {}; i < static_cast<int>(std_devs.size()); ++i) {
        const double w { 1.0 / std_devs[i] };
        x_mean += w * x_data[i];
        y_mean += w * y_data[i];
        w_sum += w;
    }
    x_mean /= w_sum;
    y_mean /= w_sum;
    a = 0.0;
    double denominator {};
    for (int i {}; i < n_data; ++i) {
        const double w { 1.0 / std_devs[i] };
        const double x_diff { (x_data[i] - x_mean) };
        a += w * x_diff * (y_data[i] - y_mean);
        denominator += w * x_diff * x_diff;
    }
    a /= denominator;
    b = y_mean - a * x_mean;
}

auto fallback_dpptrf(const int dimension, std::vector<double>& A) -> void
{
    std::vector<double> L(A.size(), 0.0);
    for (int i {}; i < dimension; ++i) {
        // L_ij
        for (int j {}; j < i; ++j) {
            const int ij { i * dimension + j };
            L[ij] = A[ij];
            for (int k {}; k < j; ++k) {
                L[ij] -= L[i * dimension + k] * L[j * dimension + k];
            }
            L[ij] = L[ij] / L[j * dimension + j];
        }
        // L_jj
        const int ii { i * dimension + i };
        L[ii] = A[ii];
        for (int k {}; k < i; ++k) {
            L[ii] -= L[i * dimension + k] * L[i * dimension + k];
        }
        L[ii] = std::sqrt(L[ii]);
    }
    A = L;
}

auto fallback_dpptrs(const int dimension,
                     const std::vector<double>& A,
                     std::vector<double>& B) -> void
{
    // Forward substitution (solving A*y=b)
    std::vector<double> y(B.size());
    for (int i {}; i < dimension; ++i) {
        y[i] = B[i];
        for (int j {}; j < i; ++j) {
            y[i] -= A[i * dimension + j] * y[j];
        }
        y[i] /= A[i * dimension + i];
    }
    // Backward substitution (solving A^T*x=y)
    for (int i { dimension - 1 }; i >= 0; --i) {
        B[i] = y[i]; // Here B=x
        for (int j { dimension - 1 }; j > i; --j) {
            // No need to explicitly use A^T here
            B[i] -= A[j * dimension + i] * B[j];
        }
        B[i] /= A[i * dimension + i];
    }
}

auto fallback_dpotri(const int dimension, std::vector<double>& A) -> void
{
    // Populate upper triangle
    for (int j {}; j < dimension; ++j) {
        for (int i {}; i < j; ++i) {
            A[i * dimension + j] = A[j * dimension + i];
        }
    }
    std::vector<double> B(dimension * dimension, 0.0);
    for (int j {}; j < dimension; ++j) {
        std::vector<double> B_slice(dimension, 0.0);
        B_slice[j] = 1.0;
        fallback_dpptrs(dimension, A, B_slice);
        for (int i {}; i < dimension; ++i) {
            B[i * dimension + j] = B_slice[i];
        }
    }
    A = B;
}

} // namespace spex
