// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

// Linear algebra routines

#pragma once

#include "packed_matrix.h"

extern "C"
{
    auto dgttrf_(const int*, double*, double*, double*, double*, int*, int*)
      -> void;
    auto dgttrs_(const char*,
                 const int*,
                 const int*,
                 const double*,
                 const double*,
                 const double*,
                 const double*,
                 const int*,
                 double*,
                 const int*,
                 int*) -> void;
    auto dppsv_(const char*,
                const int*,
                const int*,
                double*,
                double*,
                const int*,
                int*) -> void;
    auto ddot_(const int*, const double*, const int*, const double*, const int*)
      -> double;
}

namespace spex {

// Compute the solution to a real system of linear equations A * X =
// B, where A is an N-by-N symmetric positive definite matrix stored
// in packed format and X and B are N-by-NRHS matrices.
auto dppsv(const std::vector<double>& data,
           const std::vector<double>& std_devs,
           std::vector<double>& B,
           std::vector<double>& state_vector) -> void;

// Same but the B matrix is stored in packed format
auto dppsv(const std::vector<double>& data,
           const std::vector<double>& std_devs,
           PackedMatrix& B,
           std::vector<double>& state_vector) -> void;

// Solve y = a * x + b for a and b where
//
//     Sum_i (x_i - x_mean) * (y_i - y_mean)
// a = -------------------------------------
//            Sum_i (x_i - x_mean)^2
//
// b = y_mean - b * x_mean
auto simpleLinearRegression(const std::vector<double>& x_data,
                            const std::vector<double>& y_data,
                            double& a,
                            double& b) -> void;

// Same but weighted.
//
//     Sum_i w_i (x_i - x_mean) * (y_i - y_mean)
// a = -----------------------------------------
//            Sum_i w_i (x_i - x_mean)^2
//
// b = y_mean - b * x_mean
//
// x_mean = (Sum_i w_i x_1) / (Sum_i w_i)
// y_mean = (Sum_i w_i y_1) / (Sum_i w_i)
//
// where w[i] (weight) is the inverse of a standard deviation std_devs[i]
auto simpleLinearRegression(const std::vector<double>& std_devs,
                            const std::vector<double>& x_data,
                            const std::vector<double>& y_data,
                            double& a,
                            double& b) -> void;

// Fallback implementations of some of Lapack functions. Use these if,
// e.g., the linear algebra library is not thread-safe.
auto fallback_dpptrf(const int dimension, std::vector<double>& A) -> void;
auto fallback_dpptrs(const int dimension,
                     const std::vector<double>& A,
                     std::vector<double>& B) -> void;
auto fallback_dpotri(const int dimension, std::vector<double>& A) -> void;

} // namespace spex
