// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

// Class for storing all user defined configuration parameters for
// SPEXone. A single instance is used throughout the code.
//
// When adding a new parameter two places need to be updated: i)
// define the parameter by editing one of structs below or by defining
// a new struct and ii) call the scan function on the parameter in
// scanKeys so it can be read from the YAML file and/or printed on the
// screen. You may optionally add checks in Settings::checkParameters
// for things like correct file paths.

#pragma once

#include "settings.h"

namespace spex {

class SPEXSettings : public Settings
{
private:
    // Info fields for some of the parameters below. These are used
    // for multiple settings.
    static const std::string l1a_files_text;
    static const std::string l1a_file_list_text;
    static const std::string l1x_raw_text;
    static const std::string l1x_dark_text;
    static const std::string l1x_noise_text;
    static const std::string l1x_nonlin_text;
    static const std::string l1x_prnu_text;
    static const std::string l1x_stray_text;
    static const std::string l1x_fov_text;
    static const std::string l1x_rad_text;
    static const std::string dark_cols_left_text;
    static const std::string dark_cols_right_text;
    static const std::string l1a_bgr_files_text;
    static const std::string l1a_bgr_file_list_text;
    static const std::string l1a_type_text;
    static const std::string discard_frames_front_text;
    static const std::string discard_frames_back_text;
    static const std::string van_cittert_steps_text;
    static const std::string ckd_file_in_text;

    auto checkParameters() -> void override;

public:
    // Main (driver) section of configuration
    struct
    {
        Setting<std::vector<ProcLevel>> processes {
            { "main", "processes" },
            {},
            "Processes to be executed. Allowed values: dark, noise, nonlin,\n"
            "prnu, stray, fov, swath, wave, rad, pol, l1b, l1c."
        };
        Setting<std::string> ckd_file_in {
            { "main", "ckd_file_in" },
            {},
            "Input CKD file name. For steps not listed in 'processes', if the\n"
            "corresponding section is present in the input CKD file, the data\n"
            "is taken from that file and copied to the output CKD file.\n"
            "Needed for L1B and optionally CKD generation."
        };
        Setting<std::string> ckd_file_out {
            { "main", "ckd_file_out" },
            {},
            "output CKD file name (mandatory for CKD generation)"
        };
        Setting<int> detector_spat {
            { "main", "detector_spat" },
            2048,
            "number of detector pixels in the spatial dimension"
        };
        Setting<int> detector_spec {
            { "main", "detector_spec" },
            2048,
            "number of detector pixels in the spectral dimension"
        };
        Setting<std::vector<int>> bin {
            { "main", "bin" },
            { { 1, 1 } },
            "bin all measurements by these binning factors"
        };
        Setting<int> n_viewports { { "main", "n_viewports" },
                                   5,
                                   "Number of viewports" };
        Setting<std::string> binningtable {
            { "main", "binningtable" },
            {},
            "File containing binning tables, count arrays, and the lineskip\n"
            "arrays. Which set to use is determined by\n"
            "/image_attributes/binning_table from the input file(s)"
        };
        Setting<bool> spectral_binning {
            { "main", "spectral_binning" },
            true,
            "whether to spectrally bin the FOV and higher CKDs",
            SettingLevel::debug
        };
        Setting<bool> debug {
            { "main", "debug" },
            false,
            "Run all processes in debug mode, which saves additional\n"
            "information in the generated product (CKD or an L1B/L1C product)\n"
            "and produces more verbose screen output. This setting can be\n"
            "overwritten by the debug setting of individual sections",
            SettingLevel::debug
        };
        Setting<std::optional<double>> target_viewport_d {
            { "main", "target_viewport" },
            "if set, only generate CKD for this viewport, given in degrees",
            SettingLevel::debug
        };
        std::optional<int> target_viewport_i {};
        Setting<bool> use_jump_row {
            { "main", "use_jump_row" },
            true,
            "Whether to use the jump row. If set to 'false' or 'no' then any\n"
            "process that uses the dark CKD (from noise CKD generation to\n"
            "L1B) will only use the long exposure time offset.",
            SettingLevel::debug
        };
        Setting<std::vector<double>> ro_offset {
            { "main", "ro_offset" },
            {},
            "c0 and c1 parameters of the readout order dependent offset model"
        };
        Setting<std::string> isrf_file {
            { "main", "isrf_file" },
            "",
            "NetCDF file containing an ISRF database - the parameters a, m,\n"
            "s, and w_b for each ACT angle and wavelength"
        };
        Setting<std::string> solar_spectrum_file {
            { "main", "solar_spectrum_file" },
            {},
            "File containing the solar spectrum. Wavelengths and the\n"
            "spectrum are expected to be in the variables 'Vacuum Wavelength'\n"
            "and SSI. The spectrum should have the units W m-2 nm-1."
        };
        Setting<std::vector<int>> bad_rows {
            { "main", "bad_rows" },
            { 0, 1, 2046, 2047 },
            "List of bad detector rows. This is used to initialize the pixel\n"
            "mask."
        };
        Setting<double> spectrum_width {
            { "main", "spectrum_width" },
            1.0,
            "width of each spectrum, in pixels, extracted from the detector"
        };
    } main;

    // Dark offset and current calibration
    struct
    {
        Setting<double> short_exposure_time {
            { "dark", "short_exposure_time" },
            0.000324,
            "Exposure time for selecting measurements for generating the\n"
            "short exposure dark offset in seconds. Measurements with\n"
            "exposure times less than this will be used."
        };
        Setting<double> long_exposure_time {
            { "dark", "long_exposure_time" },
            0.212,
            "Exposure time for selecting measurements for generating the\n"
            "long exposure dark offset and current in seconds. Measurements\n"
            "with exposure times mores than this will be used."
        };
        Setting<std::optional<double>> last_exposure_time {
            { "dark", "last_exposure_time" },
            "Longest exposure time used in the generation of the offset and\n"
            "dark current. If not given, the number of data points for each\n"
            "linear fit is determined by the set exposure times that yield\n"
            "the lowest uncertainty of dark current."
        };
        Setting<double> data_start_fraction {
            { "dark", "data_start_fraction" },
            0.83,
            "At least this many data points are included in each linear fit.\n"
            "A higher fraction means that certain exposure times are always\n"
            "included. Only relevant for programmatic data selection and\n"
            "ignored if last_exposure_time is given."
        };
        Setting<std::optional<double>> mask_offset_min {
            { "dark", "mask_offset_min" },
            "minimum fixed dark signal (independent of integration time)"
        };
        Setting<std::optional<double>> mask_offset_max {
            { "dark", "mask_offset_max" },
            "maximum fixed dark signal (independent of integration time)"
        };
        Setting<std::optional<double>> mask_current_min {
            { "dark", "mask_current_min" },
            "minimum dark signal added per second of integration time"
        };
        Setting<std::optional<double>> mask_current_max {
            { "dark", "mask_current_max" },
            "maximum dark signal added per second of integration time"
        };
    } dark;

    // Noise calibration
    struct
    {
        Setting<std::optional<double>> mask_g_min {
            { "noise", "mask_g_min" },
            "Pixels whose inverse noise gain (g) is lower than this value are\n"
            "masked bad."
        };
        Setting<std::optional<double>> mask_g_max {
            { "noise", "mask_g_max" },
            "Pixels whose inverse noise gain (g) is higher than this value\n"
            "are masked bad."
        };
        Setting<std::optional<double>> mask_n_min {
            { "noise", "mask_n_min" },
            "Pixels whose dark noise (n) is lower than this value are masked\n"
            "bad."
        };
        Setting<std::optional<double>> mask_n_max {
            { "noise", "mask_n_max" },
            "Pixels whose dark noise (n) is higher than this value are masked\n"
            "bad."
        };
        Setting<double> intensity_threshold {
            { "noise", "intensity_threshold" },
            730.0,
            "Signal intensities higher than this value are excluded from the\n"
            "noise model fit."
        };
        struct
        {
            Setting<int> iteration_limit {
                { "noise", "fitting", "iteration_limit" },
                100,
                "iteration limit",
                SettingLevel::debug
            };
            Setting<double> initial_g {
                { "noise", "fitting", "initial_g" },
                0.08,
                "initial value for inverse noise gain (g)",
                SettingLevel::debug
            };
            Setting<double> initial_n { { "noise", "fitting", "initial_n" },
                                        3.0,
                                        "initial value for dark noise (n)",
                                        SettingLevel::debug };
        } fitting;
    } noise;

    // Nonlinearity calibration
    struct
    {
        Setting<NonlinModel> model {
            { "nonlin", "model" },
            NonlinModel::sigmoidal,
            "Model used for deriving the nonlinearity CKD. Allowed values:\n"
            "sigmoidal, bspline, single"
        };
        Setting<double> target_exposure_time {
            { "nonlin", "target_exposure_time" },
            0.066,
            "exposure time for which to generate the nonlinearity CKD"
        };
        Setting<int> target_counts_min {
            { "nonlin", "target_counts_min" },
            -5,
            "lowest value of the expected measured signal (in single\n"
            "exposure time mode this is the lower bounds of the linear domain"
        };
        Setting<int> target_counts_max {
            { "nonlin", "target_counts_max" },
            900,
            "highest value of the expected measured signal (in single\n"
            "exposure time mode this is the upper bounds of the linear domain"
        };
        Setting<int> n_knots {
            { "nonlin", "n_knots" },
            20,
            "Number of spline knots. This affects storage and memory\n"
            "footprint but does not affect evaluation speed which is O(1)."
        };
        Setting<std::string> restart_ckd {
            { "nonlin", "restart_ckd" },
            "",
            "If given, restart the CKD generation from an earlier CKD and\n"
            "skip the time consuming steps. This is useful if you only want\n"
            "to change the number knots and their range but not redo the full\n"
            "derivation. In order for this to work the input CKD needs to\n"
            "contain nonlinearity debug information."
        };
        Setting<bool> scale_median {
            { "nonlin", "scale_median" },
            true,
            "Flag to scale the spline knots with the median of the saturated\n"
            "image. Disable to use individual-pixel saturation levels."
        };
        Setting<std::vector<double>> knots {
            { "nonlin", "knots" },
            {},
            "B-spline knots. If given, overrides target_counts_min,\n"
            "target_counts_max, and n_knots."
        };
        Setting<int> order { { "nonlin", "order" }, 3, "B-spline order" };
        Setting<std::optional<double>> mask_chi2_max {
            { "nonlin", "mask_chi2_max" },
            "maximum residual chi squared on non-linearity correction"
        };
        Setting<std::vector<int>> debug_pixels {
            { "nonlin", "debug_pixels" },
            {},
            "if in debug mode then include additional data about these\n"
            "pixels in the CKD (no effect in non-debug mode)"
        };
        Setting<bool> use_all_signal_rates {
            { "nonlin", "use_all_signal_rates" },
            true,
            "Whether to use all signal rates when computing the reference\n"
            "diode f1 value. Alternative is to to compute an average signal\n"
            "rate per light intensity and use that in the second step when\n"
            "fitting for f1."
        };
        Setting<bool> include_f2 {
            { "nonlin", "include_f2" },
            true,
            "whether to include both f1 and f2 parameters when fitting for\n"
            "the linear signal rates or only f1"
        };
        Setting<double> sqrt_factor {
            { "nonlin", "sqrt_factor" },
            -0.17,
            "Factor X in the offset model X*sqrt(S) where S is signal. The\n"
            "quantity X*sqrt(S) is subracted from the signal, meaning for a\n"
            "negative X value the pixel light levels are increased.",
            SettingLevel::debug
        };
    } nonlin;

    // PRNU calibration
    struct
    {
        Setting<int> order_spat {
            { "prnu", "order_spat" },
            0,
            "polynomial order for the spatial dimension"
        };
        Setting<int> order_spec {
            { "prnu", "order_spec" },
            0,
            "polynomial order for the spectral dimension"
        };
        Setting<std::optional<double>> mask_prnu_min {
            { "prnu", "mask_prnu_min" },
            "minimum pixel response non-uniformity"
        };
        Setting<std::optional<double>> mask_prnu_max {
            { "prnu", "mask_prnu_max" },
            "maximum pixel response non-uniformity"
        };
        struct
        {
            Setting<int> iteration_limit {
                { "prnu", "fitting", "iteration_limit" },
                100,
                "iteration limit",
                SettingLevel::debug
            };
        } fitting;
    } prnu;

    // Stray light calibration
    struct
    {
        Setting<double> exposure_intensity_thres {
            { "stray", "exposure_intensity_thres" },
            730.0,
            "For pixel light levels higher than this threshold, switch to an\n"
            "image with a longer exposure time instead. This parameter\n"
            "affects how composite images are built up from multiple exposure\n"
            "times."
        };
        Setting<double> kernel_hole_radius {
            { "stray", "kernel_hole_radius" },
            10.0,
            "radius of the center hole when constructing the kernels"
        };
        Setting<int> kernel_extent {
            { "stray", "kernel_extent" },
            512,
            "Kernels are confined to a square box with the side\n"
            "2*kernel_extent. Beyond that the kernel is taken to be zero."
        };
        Setting<bool> reduced_kernels {
            { "stray", "reduced_kernels" },
            false,
            "If true, kernels are reduced in size to\n"
            "detector_spec/2 * detector_spat/2, e.g. 1024 * 1024. Otherwise\n"
            "the kernels have the same dimensions as the detector.\n"
            "Implication is that for deconvolution the images don't have to\n"
            "be fully unbinned which reduces computational cost."
        };
        struct
        {
            Setting<bool> write {
                { "stray", "checkpoint", "write" },
                false,
                "Whether to write the raw kernels and their centers to a\n"
                "checkpoint file. They can be later read by setting\n"
                "[stray][checkpoint][read] to yes.",
                SettingLevel::debug
            };
            Setting<bool> read {
                { "stray", "checkpoint", "read" },
                false,
                "whether to skip the calibration and construction of raw\n"
                "kernels and their centers and read them directly from a\n"
                "checkpoint file",
                SettingLevel::debug
            };
            Setting<std::string> filename {
                { "stray", "checkpoint", "filename" },
                {},
                "name of the checkpoint file for reading or writing the raw\n"
                "kernels and their centers",
                SettingLevel::debug
            };
        } checkpoint;
    } stray;

    // FOV calibration
    struct
    {
        Setting<int> n_knots {
            { "fov", "n_knots" },
            2,
            "Number of B-spline knots when smoothing the FOV curves"
        };
        Setting<int> bspline_order {
            { "fov", "bspline_order" },
            3,
            "B-spline (polynomial) order when smoothing the FOV curves"
        };
        Setting<int> exclude_columns_left {
            { "fov", "exclude_columns_left" },
            {},
            "Number of columns from the left side of the detector to be\n"
            "excluded when determining peaks. The edge columns have no useful\n"
            "signal which can lead to instabilities when trying to fit with a\n"
            "Gaussian."
        };
        Setting<int> exclude_columns_right {
            { "fov", "exclude_columns_right" },
            {},
            "Same as exclude_columns_left but applies to the right edge (long\n"
            "wavelengths) of the detector."
        };
        Setting<std::vector<double>> act_sampling {
            { "fov", "act_sampling" },
            {},
            "ACT angle sampling per viewport in degrees"
        };
        struct
        {
            Setting<int> iteration_limit {
                { "fov", "fitting", "iteration_limit" },
                30,
                "iteration limit for finding peaks",
                SettingLevel::debug
            };
        } fitting;
    } fov;

    // Swath calibration
    struct
    {
        Setting<std::string> theodolite_reference_filename {
            { "swath", "theodolite_reference_filename" },
            {},
            "NetCDF file name of the theodolite reference file (a lot of\n"
            "measurements on the wrong day)"
        };
        Setting<std::string> theodolite_spotcheck_filename {
            { "swath", "theodolite_spotcheck_filename" },
            {},
            "NetCDF file name of the theodolite spot-check file (one\n"
            "measurement on the correct day)"
        };
        Setting<std::vector<double>> viewport_alt_angles {
            { "swath", "viewport_alt_angles" },
            {},
            "Expected ALT angles of each viewport (also used to put the\n"
            "viewports in the correct order)"
        };
        Setting<int> order {
            { "swath", "order" },
            {},
            "Polynomial order for fitting ALT against ACT angles"
        };
        // Read the rotation matrix as a vector and then convert to a 3x3 array
        Setting<std::vector<double>> rotmat_in {
            { "swath", "rotmat" },
            { { 1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0 } },
            "rotation matrix from spacecraft to instrument coordinates"
        };
        std::array<double, dims::vec * dims::vec> rotmat {};
        struct
        {
            Setting<bool> write {
                { "swath", "checkpoint", "write" },
                false,
                "Whether to write the computed ACT and ALT angles to a\n"
                "checkpoint file. They can read by setting\n"
                "[swath][checkpoint][read] to yes.",
                SettingLevel::debug
            };
            Setting<bool> read {
                { "swath", "checkpoint", "read" },
                false,
                "whether to skip the computation of ACT and ALT angles "
                "and\n"
                "read them directly from a checkpoint file",
                SettingLevel::debug
            };
            Setting<std::string> filename {
                { "swath", "checkpoint", "filename" },
                {},
                "name of the checkpoint file for reading or writing the\n"
                "checkpoint data",
                SettingLevel::debug
            };
        } checkpoint;
    } swath;

    // Spectral calibration
    struct
    {
        Setting<int> order {
            { "wave", "order" },
            0,
            "Polynomial order for fitting column indices vs wavelengths. The\n"
            "fitted polynomial is then used to generated the wave_full\n"
            "variable found in the CKD."
        };
        Setting<double> peak_offset {
            { "wave", "peak_offset" },
            {},
            "Offset added to all fitted peaks found in the reference spectrum",
        };
        Setting<std::vector<double>> reference_wavelengths {
            { "wave", "reference_wavelengths" },
            {},
            "Values to replace the peaks found in the references spectrum.\n"
            "Run the wavelength generation in debug mode to be sure which\n"
            "peaks are used."
        };
    } wave;

    // Polarimetric calibration
    struct
    {
        Setting<std::string> delta_prior_file {
            { "pol", "delta_prior_file" },
            {},
            "File containing prior data for retardance"
        };
        Setting<std::string> mt_prior_file {
            { "pol", "mt_prior_file" },
            {},
            "File containing prior data for telescope polarization parameter mt"
        };
        Setting<std::optional<EGSEMember>> aolp_from {
            { "pol", "aolp_from" },
            "Name of the EGSE member from which to read the AoLP values.\n"
            "Allowed values: POLARIZER, ACT_ANGLE"
        };
        Setting<std::optional<double>> aolp_offset {
            { "pol", "aolp_offset" },
            "Value in degrees added to each "
            "polarizer angle read from the egse\n"
            "variable"
        };
        Setting<std::optional<bool>> aolp_sign_flip {
            { "pol", "aolp_sign_flip" },
            "Whether the polarizer angles read from the egse variable are\n"
            "interpreted as having the opposite sign"
        };
        Setting<std::optional<double>> fit_window_min {
            { "pol", "fit_window_min" },
            "Shortest wavelength in nm included in the fitting procedure.\n"
            "This parameter could be useful in excluding bad data points near\n"
            "the edge of the detector that could have an adverse effect on\n"
            "the fitting procedure.",
            SettingLevel::debug
        };
        Setting<std::optional<double>> fit_window_max {
            { "pol", "fit_window_max" },
            "Longest wavelength in nm included in the fitting procedure. This\n"
            "parameter could be useful in excluding bad data points near the\n"
            "edge of the detector that could have an adverse effect on the\n"
            "fitting procedure.",
            SettingLevel::debug
        };
        struct
        {
            Setting<int> delta_order {
                { "pol", "fitting", "delta_order" },
                4,
                "Polynomial order of retardance used in the model",
            };
            Setting<int> eps_order {
                { "pol", "fitting", "eps_order" },
                4,
                "Polynomial order of efficiency used in the model",
            };
            // Internally, the default value is -1
            Setting<std::optional<int>> mt_order {
                { "pol", "fitting", "mt_order" },
                "Polynomial order of telescope polarization Mueller element\n"
                "used in the model. Default is to not fit for mt.",
            };
        } fitting;
        Setting<int> isrf_cutoff {
            { "pol", "isrf_cutoff" },
            500,
            "Extent of the ISRF in pixels. Lowering this value makes the\n"
            "integration run faster when generating the solar spectrum\n"
            "correction (Fraunhofer) factors.",
            SettingLevel::debug
        };
        struct
        {
            Setting<bool> enabled {
                { "pol", "mueller", "enabled" },
                false,
                "whether to generate the polarimetric CKD using the Mueller\n"
                "matrix approach"
            };
            Setting<std::optional<int>> knots {
                { "pol", "mueller", "knots" },
                "if present, this many knots will be used in B-spline\n"
                "smoothing of the S and P spectra before applying the Mueller\n"
                "approach"
            };
            Setting<int> b_spline_order {
                { "pol", "mueller", "b_spline_order" },
                3,
                "B-spline order for smoothing the S and P spectra (only used\n"
                "if 'knots' is given)"
            };
        } mueller;
        struct
        {
            Setting<bool> enabled {
                { "pol", "print_debug", "enabled" },
                false,
                "Whether to print the modulation data and the corresponding\n"
                "fits for each value of AoLP of a given viewport and field of\n"
                "view (FOV). Also printed are values of the fitting\n"
                "parameters.",
                SettingLevel::debug
            };
            Setting<int> viewport {
                { "pol", "print_debug", "viewport" },
                {},
                "Viewport for which to print the fitting results. Counting\n"
                "starts from 0. Ignored if debug_fov < 0.",
                SettingLevel::debug
            };
            Setting<int> fov {
                { "pol", "print_debug", "fov" },
                {},
                "Field of view for which to print the fitting results.\n"
                "Counting starts from 0 and is relative to the edge of the\n"
                "given viewport. Ignored if debug_vp < 0.",
                SettingLevel::debug
            };
            Setting<std::string> output_folder {
                { "pol", "print_debug", "output_folder" },
                {},
                "Full or relative path to the folder that will contain the\n"
                "fitting results. Only used with debug_vp and debug_fov.\n"
                "Default is to use the same directory from which the\n"
                "executable is run.",
                SettingLevel::debug
            };
        } print_debug;
    } pol;

    struct
    {
        Setting<std::string> outputfile { { "l1b", "outputfile" },
                                          {},
                                          "Output L1B file" };
        Setting<int> order {
            { "l1b", "order" },
            {},
            "B-spline order of wavelength dependence of q and u"
        };
        Setting<std::vector<double>> knots {
            { "l1b", "knots" },
            {},
            "Wavelengths of knot positions in B-spline"
        };
        Setting<int> n_knots {
            { "l1b", "n_knots" },
            {},
            "Number of B-spline knots. If given, only the first and last\n"
            "value of [l1b][knots] are taken and the intermediate values are\n"
            "generated automatically. This is a convenience parameter.\n"
            "Alternatively, [l1b][knots] can be used to define the whole knot\n"
            "range by hand."
        };
        Setting<std::vector<double>> l1b_wavelength {
            { "l1b", "l1b_wavelength" },
            {},
            "Central wavelengths of polarimetric product"
        };
        Setting<int> n_l1b_wavelength {
            { "l1b", "n_l1b_wavelength" },
            {},
            "If given, defines the number of L1B wavelengths. The first and\n"
            "last value of [l1b][l1b_wavelength] define the wavelength range\n"
            "and all intermediate values are generated based on\n"
            "[l1b][n_l1b_wavelength]. Works similarly as [l1b][knots]."
        };
        Setting<std::vector<double>> intensity_wavelength {
            { "l1b", "intensity_wavelength" },
            {},
            "If given, defines the number of intensity grid wavelengths.\n"
            "Works the same way as [l1b][n_knots] or [l1b][n_l1b_wavelength]."
        };
        Setting<int> n_intensity_wavelength {
            { "l1b", "n_intensity_wavelength" },
            {},
            "Wavelengths for intensity-only grid"
        };
        Setting<double> resolving_power {
            { "l1b", "resolving_power" },
            {},
            "Full-width half-maximum of Gauss convolution of polarimetric L1B\n"
            "product"
        };
        Setting<double> gauss_range_tol {
            { "l1b", "gauss_range_tol" },
            1e-5,
            "Values higher than this are considered non-zero in the Gaussian\n"
            "convolution kernel. Zero values are not present in the packed\n"
            "matrix format which speeds up the calculation. This affects the\n"
            "accuracy of the convolved I, Q, U and related quantities."
        };
        Setting<bool> geolocation {
            { "l1b", "geolocation" },
            true,
            "Wether to perform geolocation (disable when using OCAL\n"
            "measurements as test)",
        };
        Setting<std::string> dem_file {
            { "l1b", "dem_file" },
            {},
            "Digital elevation model. Leave empty for placeholder DEM function."
        };
        Setting<std::optional<ProcLevel>> last_calibration_step {
            { "l1b", "last_calibration_step" },
            "If given, the last calibration step to be executed. Allowed\n"
            "values: dark, noise, nonlin, prnu, stray, fov, swath, wave, rad.",
            SettingLevel::debug
        };
        Setting<int> image_start {
            { "l1b", "image_start" },
            {},
            "first image to be processed (counting starts at 0)",
            SettingLevel::debug
        };
        Setting<std::optional<int>> image_end {
            { "l1b", "image_end" },
            "last image to be processed (up to the number of images minus 1)",
            SettingLevel::debug
        };
        Setting<bool> noise_propagation {
            { "l1b", "noise_propagation" },
            true,
            "Whether to propagate noise through demodulation or turn it off\n"
            "to save time. Note that the noise is still included in the L1B\n"
            "product but with incorrect values.",
            SettingLevel::debug
        };
        Setting<bool> noise_propagation_telescope {
            { "l1b", "noise_propagation_telescope" },
            false,
            "whether to propagate noise that is associated with telescope\n"
            "polarization through demodulation or turn it off",
            SettingLevel::debug
        };
        Setting<bool> do_fraunhofer {
            { "l1b", "do_fraunhofer" },
            true,
            "Whether to apply the solar spectrum correction (Fraunhofer)"
        };
        Setting<bool> dem_in_memory {
            { "l1b", "dem_in_memory" },
            true,
            "where to load the full digital elevation model into memory"
        };
        Setting<bool> l1a_in_memory {
            { "l1b", "l1a_in_memory" },
            false,
            "where to preload all L1A detector images to memory"
        };
        Setting<bool> bgr_matching {
            { "l1b", "bgr_matching" },
            true,
            "Whether to automatically match a background to a measurement.\n"
            "The matching algorithm depends on the process but most common is\n"
            "to match by viewport. If disabled, L1A products in l1a_bgr_files\n"
            "are matched to those in l1a_files one by one in the order they\n"
            "appear",
            SettingLevel::debug
        };
        Setting<bool> fallback_linear_algebra {
            { "l1b", "fallback_linear_algebra" },
            true,
            "whether to use fallback linear algebra functions instead of a\n"
            "linear algebra library"
        };
        Setting<std::vector<double>> rotmat_in {
            { "l1b", "rotmat" },
            { { 1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0 } },
            "Rotation matrix applied to swath vectors in the instrument\n"
            "reference frame. Can be used for additional adjustments."
        };
        std::array<double, dims::vec * dims::vec> rotmat {};
    } l1b;

    struct
    {
        Setting<std::string> l1b_in {
            { "l1c", "l1b_in" },
            {},
            "Input L1B file (no L1A files are used here)"
        };
        Setting<std::string> l1c_in { { "l1c", "l1c_in" },
                                      {},
                                      "L1C grid file as input" };
        Setting<std::string> l1c_out { { "l1c", "l1c_out" },
                                       {},
                                       "L1C product" };
        Setting<std::string> dem_file { { "l1c", "dem_file" },
                                        {},
                                        "digital elevation model" };
        Setting<std::vector<int>> cols_to_nadir {
            { "l1c", "cols_to_nadir" },
            { 15, 14 },
            "number of columns to the left and right of the nadir bin that\n"
            "the L1C grid is reduced to"
        };
    } l1c;

    // Configuration parameters used by most processes. For simplicity
    // and consistency, all parameter lists have the same length which
    // is equal to the number of process levels (see ProcLevel). Some
    // parameters are thus unused (or undefined for now). For
    // instance, even though l1a_files spans over all processes, no
    // L1A files are currently used as input for the L1C process.
    std::array<Setting<std::vector<std::string>>,
               static_cast<int>(ProcLevel::n_levels)>
      l1a_files { {
        {}, // l1a
        {}, // raw
        { { "dark", "l1a_files" }, {}, l1a_files_text + "dark CKD" },
        { { "noise", "l1a_files" }, {}, l1a_files_text + "noise CKD" },
        { { "nonlin", "l1a_files" }, {}, l1a_files_text + "nonlinearity CKD" },
        { { "prnu", "l1a_files" }, {}, l1a_files_text + "PRNU CKD" },
        { { "stray", "l1a_files" }, {}, l1a_files_text + "stray light CKD" },
        { { "fov", "l1a_files" }, {}, l1a_files_text + "field of view CKD" },
        { { "swath", "l1a_files" }, {}, l1a_files_text + "swath vector CKD" },
        { { "wave", "l1a_files" }, {}, l1a_files_text + "spectral CKD" },
        { { "rad", "l1a_files" }, {}, l1a_files_text + "radiometric CKD" },
        { { "pol", "l1a_files" }, {}, l1a_files_text + "polarimetric CKD" },
        { { "l1b", "l1a_files" }, {}, l1a_files_text + "L1B product" },
        {} // l1c
      } };
    std::array<Setting<std::string>, static_cast<int>(ProcLevel::n_levels)>
      l1a_file_list { {
        {}, // l1a
        {}, // raw
        { { "dark", "l1a_file_list" }, {}, l1a_file_list_text + "dark CKD" },
        { { "noise", "l1a_file_list" }, {}, l1a_file_list_text + "noise CKD" },
        { { "nonlin", "l1a_file_list" },
          {},
          l1a_file_list_text + "nonlinearity CKD" },
        { { "prnu", "l1a_file_list" }, {}, l1a_file_list_text + "PRNU CKD" },
        { { "stray", "l1a_file_list" },
          {},
          l1a_file_list_text + "stray light CKD" },
        { { "fov", "l1a_file_list" },
          {},
          l1a_file_list_text + "field of view CKD" },
        { { "swath", "l1a_file_list" },
          {},
          l1a_file_list_text + "swath vector CKD" },
        { { "wave", "l1a_file_list" },
          {},
          l1a_file_list_text + "spectral CKD" },
        { { "rad", "l1a_file_list" },
          {},
          l1a_file_list_text + "radiometric CKD" },
        { { "pol", "l1a_file_list" },
          {},
          l1a_file_list_text + "polarimetric CKD" },
        { { "l1b", "l1a_file_list" }, {}, l1a_file_list_text + "L1B product" },
        {} // l1c
      } };
    std::array<Setting<std::string>, static_cast<int>(ProcLevel::n_levels)>
      l1x_raw { {
        {}, // l1a
        {}, // raw
        { { "dark", "l1x_raw" }, {}, l1x_raw_text, SettingLevel::debug },
        { { "noise", "l1x_raw" }, {}, l1x_raw_text, SettingLevel::debug },
        { { "nonlin", "l1x_raw" }, {}, l1x_raw_text, SettingLevel::debug },
        { { "prnu", "l1x_raw" }, {}, l1x_raw_text, SettingLevel::debug },
        { { "stray", "l1x_raw" }, {}, l1x_raw_text, SettingLevel::debug },
        { { "fov", "l1x_raw" }, {}, l1x_raw_text, SettingLevel::debug },
        { { "swath", "l1x_raw" }, {}, l1x_raw_text, SettingLevel::debug },
        { { "wave", "l1x_raw" }, {}, l1x_raw_text, SettingLevel::debug },
        { { "rad", "l1x_raw" }, {}, l1x_raw_text, SettingLevel::debug },
        { { "pol", "l1x_raw" }, {}, l1x_raw_text, SettingLevel::debug },
        { { "l1b", "l1x_raw" }, {}, l1x_raw_text, SettingLevel::debug },
        {} // l1c
      } };
    std::array<Setting<std::string>, static_cast<int>(ProcLevel::n_levels)>
      l1x_dark { {
        {}, // l1a
        {}, // raw
        {}, // dark
        { { "noise", "l1x_dark" }, {}, l1x_dark_text, SettingLevel::debug },
        { { "nonlin", "l1x_dark" }, {}, l1x_dark_text, SettingLevel::debug },
        { { "prnu", "l1x_dark" }, {}, l1x_dark_text, SettingLevel::debug },
        { { "stray", "l1x_dark" }, {}, l1x_dark_text, SettingLevel::debug },
        { { "fov", "l1x_dark" }, {}, l1x_dark_text, SettingLevel::debug },
        { { "swath", "l1x_dark" }, {}, l1x_dark_text, SettingLevel::debug },
        { { "wave", "l1x_dark" }, {}, l1x_dark_text, SettingLevel::debug },
        { { "rad", "l1x_dark" }, {}, l1x_dark_text, SettingLevel::debug },
        { { "pol", "l1x_dark" }, {}, l1x_dark_text, SettingLevel::debug },
        { { "l1b", "l1x_dark" }, {}, l1x_dark_text, SettingLevel::debug },
        {} // l1c
      } };
    std::array<Setting<std::string>, static_cast<int>(ProcLevel::n_levels)>
      l1x_nonlin { {
        {}, // l1a
        {}, // raw
        {}, // dark
        {}, // noise
        {}, // nonlin
        { { "prnu", "l1x_nonlin" }, {}, l1x_nonlin_text, SettingLevel::debug },
        { { "stray", "l1x_nonlin" }, {}, l1x_nonlin_text, SettingLevel::debug },
        { { "fov", "l1x_nonlin" }, {}, l1x_nonlin_text, SettingLevel::debug },
        { { "swath", "l1x_nonlin" }, {}, l1x_nonlin_text, SettingLevel::debug },
        { { "wave", "l1x_nonlin" }, {}, l1x_nonlin_text, SettingLevel::debug },
        { { "rad", "l1x_nonlin" }, {}, l1x_nonlin_text, SettingLevel::debug },
        { { "pol", "l1x_nonlin" }, {}, l1x_nonlin_text, SettingLevel::debug },
        { { "l1b", "l1x_nonlin" }, {}, l1x_nonlin_text, SettingLevel::debug },
        {} // l1c
      } };
    std::array<Setting<std::string>, static_cast<int>(ProcLevel::n_levels)>
      l1x_prnu { {
        {}, // l1a
        {}, // raw
        {}, // dark
        {}, // noise
        {}, // nonlin
        {}, // prnu
        { { "stray", "l1x_prnu" }, {}, l1x_prnu_text, SettingLevel::debug },
        { { "fov", "l1x_prnu" }, {}, l1x_prnu_text, SettingLevel::debug },
        { { "swath", "l1x_prnu" }, {}, l1x_prnu_text, SettingLevel::debug },
        { { "wave", "l1x_prnu" }, {}, l1x_prnu_text, SettingLevel::debug },
        { { "rad", "l1x_prnu" }, {}, l1x_prnu_text, SettingLevel::debug },
        { { "pol", "l1x_prnu" }, {}, l1x_prnu_text, SettingLevel::debug },
        { { "l1b", "l1x_prnu" }, {}, l1x_prnu_text, SettingLevel::debug },
        {} // l1c
      } };
    std::array<Setting<std::string>, static_cast<int>(ProcLevel::n_levels)>
      l1x_stray { {
        {}, // l1a
        {}, // raw
        {}, // dark
        {}, // noise
        {}, // nonlin
        {}, // prnu
        {}, // stray
        { { "fov", "l1x_stray" }, {}, l1x_stray_text, SettingLevel::debug },
        { { "swath", "l1x_stray" }, {}, l1x_stray_text, SettingLevel::debug },
        { { "wave", "l1x_stray" }, {}, l1x_stray_text, SettingLevel::debug },
        { { "rad", "l1x_stray" }, {}, l1x_stray_text, SettingLevel::debug },
        { { "pol", "l1x_stray" }, {}, l1x_stray_text, SettingLevel::debug },
        { { "l1b", "l1x_stray" }, {}, l1x_stray_text, SettingLevel::debug },
        {} // l1c
      } };
    std::array<Setting<std::string>, static_cast<int>(ProcLevel::n_levels)>
      l1x_fov { {
        {}, // l1a
        {}, // raw
        {}, // dark
        {}, // noise
        {}, // nonlin
        {}, // prnu
        {}, // stray
        {}, // fov
        { { "swath", "l1x_fov" }, {}, l1x_fov_text, SettingLevel::debug },
        { { "wave", "l1x_fov" }, {}, l1x_fov_text, SettingLevel::debug },
        { { "rad", "l1x_fov" }, {}, l1x_fov_text, SettingLevel::debug },
        { { "pol", "l1x_fov" }, {}, l1x_fov_text, SettingLevel::debug },
        { { "l1b", "l1x_fov" }, {}, l1x_fov_text, SettingLevel::debug },
        {} // l1c
      } };
    std::array<Setting<std::string>, static_cast<int>(ProcLevel::n_levels)>
      l1x_rad { {
        {}, // l1a
        {}, // raw
        {}, // dark
        {}, // noise
        {}, // nonlin
        {}, // prnu
        {}, // stray
        {}, // fov
        {}, // swath
        {}, // wave
        {}, // rad
        { { "pol", "l1x_rad" }, {}, l1x_rad_text, SettingLevel::debug },
        { { "l1b", "l1x_rad" }, {}, l1x_rad_text, SettingLevel::debug },
        {} // l1c
      } };
    std::array<Setting<int>, static_cast<int>(ProcLevel::n_levels)>
      dark_cols_left { {
        {}, // l1a
        {}, // raw
        {}, // dark
        {}, // noise
        {}, // nonlin
        {}, // prnu
        { { "stray", "dark_cols_left" },
          {},
          dark_cols_left_text + "stray light images only." },
        { { "fov", "dark_cols_left" },
          {},
          dark_cols_left_text + "FOV images only." },
        { { "swath", "dark_cols_left" },
          {},
          dark_cols_left_text + "swath images only." },
        { { "wave", "dark_cols_left" },
          {},
          dark_cols_left_text + "wavelength images only." },
        { { "rad", "dark_cols_left" },
          {},
          dark_cols_left_text + "radiometric images only." },
        { { "pol", "dark_cols_left" },
          {},
          dark_cols_left_text + "polarimetric images only." },
        { { "l1b", "dark_cols_left" },
          {},
          dark_cols_left_text + "the l1b process only." },
        {} // l1c
      } };
    std::array<Setting<int>, static_cast<int>(ProcLevel::n_levels)>
      dark_cols_right { {
        {}, // l1a
        {}, // raw
        {}, // dark
        {}, // noise
        {}, // nonlin
        {}, // prnu
        { { "stray", "dark_cols_right" },
          {},
          dark_cols_right_text + "stray light images only." },
        { { "fov", "dark_cols_right" },
          {},
          dark_cols_right_text + "FOV images only." },
        { { "swath", "dark_cols_right" },
          {},
          dark_cols_right_text + "swath images only." },
        { { "wave", "dark_cols_right" },
          {},
          dark_cols_right_text + "wavelength images only." },
        { { "rad", "dark_cols_right" },
          {},
          dark_cols_right_text + "radiometric images only." },
        { { "pol", "dark_cols_right" },
          {},
          dark_cols_right_text + "polarimetric images only." },
        { { "l1b", "dark_cols_right" },
          {},
          dark_cols_right_text + "the l1b process only." },
        {} // l1c
      } };
    std::array<Setting<std::vector<std::string>>,
               static_cast<int>(ProcLevel::n_levels)>
      l1a_bgr_files { {
        {}, // l1a
        {}, // raw
        {}, // dark
        {}, // noise
        { { "nonlin", "l1a_bgr_files" }, {}, l1a_bgr_files_text },
        { { "prnu", "l1a_bgr_files" }, {}, l1a_bgr_files_text },
        { { "stray", "l1a_bgr_files" }, {}, l1a_bgr_files_text },
        { { "fov", "l1a_bgr_files" }, {}, l1a_bgr_files_text },
        { { "swath", "l1a_bgr_files" }, {}, l1a_bgr_files_text },
        { { "wave", "l1a_bgr_files" }, {}, l1a_bgr_files_text },
        { { "rad", "l1a_bgr_files" }, {}, l1a_bgr_files_text },
        { { "pol", "l1a_bgr_files" }, {}, l1a_bgr_files_text },
        { { "l1b", "l1a_bgr_files" }, {}, l1a_bgr_files_text },
        {} // l1c
      } };
    std::array<Setting<std::string>, static_cast<int>(ProcLevel::n_levels)>
      l1a_bgr_file_list { {
        {}, // l1a
        {}, // raw
        {}, // dark
        {}, // noise
        { { "nonlin", "l1a_bgr_file_list" }, {}, l1a_bgr_file_list_text },
        { { "prnu", "l1a_bgr_file_list" }, {}, l1a_bgr_file_list_text },
        { { "stray", "l1a_bgr_file_list" }, {}, l1a_bgr_file_list_text },
        { { "fov", "l1a_bgr_file_list" }, {}, l1a_bgr_file_list_text },
        { { "swath", "l1a_bgr_file_list" }, {}, l1a_bgr_file_list_text },
        { { "wave", "l1a_bgr_file_list" }, {}, l1a_bgr_file_list_text },
        { { "rad", "l1a_bgr_file_list" }, {}, l1a_bgr_file_list_text },
        { { "pol", "l1a_bgr_file_list" }, {}, l1a_bgr_file_list_text },
        { { "l1b", "l1a_bgr_file_list" }, {}, l1a_bgr_file_list_text },
        {} // l1c
      } };
    std::array<Setting<L1AType>, static_cast<int>(ProcLevel::n_levels)>
      l1a_type { {
        {}, // l1a
        {}, // raw
        { { "dark", "l1a_type" },
          L1AType::diagnostic,
          l1a_type_text,
          SettingLevel::debug },
        { { "noise", "l1a_type" },
          L1AType::diagnostic,
          l1a_type_text,
          SettingLevel::debug },
        { { "nonlin", "l1a_type" },
          L1AType::diagnostic,
          l1a_type_text,
          SettingLevel::debug },
        { { "prnu", "l1a_type" },
          L1AType::diagnostic,
          l1a_type_text,
          SettingLevel::debug },
        { { "stray", "l1a_type" },
          L1AType::diagnostic,
          l1a_type_text,
          SettingLevel::debug },
        { { "fov", "l1a_type" },
          L1AType::diagnostic,
          l1a_type_text,
          SettingLevel::debug },
        { { "swath", "l1a_type" },
          L1AType::diagnostic,
          l1a_type_text,
          SettingLevel::debug },
        { { "wave", "l1a_type" },
          L1AType::diagnostic,
          l1a_type_text,
          SettingLevel::debug },
        { { "rad", "l1a_type" },
          L1AType::diagnostic,
          l1a_type_text,
          SettingLevel::debug },
        { { "pol", "l1a_type" },
          L1AType::diagnostic,
          l1a_type_text,
          SettingLevel::debug },
        { { "l1b", "l1a_type" },
          L1AType::science,
          l1a_type_text,
          SettingLevel::debug },
        {} // l1c
      } };
    std::array<Setting<int>, static_cast<int>(ProcLevel::n_levels)>
      discard_frames_front { {
        {}, // l1a
        {}, // raw
        { { "dark", "discard_frames_front" }, {}, discard_frames_front_text },
        { { "noise", "discard_frames_front" }, {}, discard_frames_front_text },
        { { "nonlin", "discard_frames_front" }, {}, discard_frames_front_text },
        {}, // prnu
        {}, // stray
        {}, // fov
        {}, // swath
        {}, // wave
        {}, // rad
        {}, // pol
        {}, // l1b
        {}, // l1c
      } };
    std::array<Setting<int>, static_cast<int>(ProcLevel::n_levels)>
      discard_frames_back { {
        {}, // l1a
        {}, // raw
        { { "dark", "discard_frames_back" }, {}, discard_frames_back_text },
        { { "noise", "discard_frames_back" }, {}, discard_frames_back_text },
        { { "nonlin", "discard_frames_back" }, {}, discard_frames_back_text },
        {}, // prnu
        {}, // stray
        {}, // fov
        {}, // swath
        {}, // wave
        {}, // rad
        {}, // pol
        {}, // l1b
        {}, // l1c
      } };
    // Make sure to call value() when evaluating these! Otherwise
    // you're evaluating std::optional itself which always returns
    // true because it is set in SPEXSettings::checkParameters.
    std::array<Setting<std::optional<bool>>,
               static_cast<int>(ProcLevel::n_levels)>
      debug { {
        {}, // l1a
        {}, // raw
        { { "dark", "debug" },
          "Run dark CKD generation in debug mode",
          SettingLevel::debug },
        { { "noise", "debug" },
          "Run noise CKD generation in debug mode",
          SettingLevel::debug },
        { { "nonlin", "debug" },
          "Run nonlinearity CKD generation in debug mode",
          SettingLevel::debug },
        { { "prnu", "debug" },
          "Run PRNU CKD generation in debug mode",
          SettingLevel::debug },
        { { "stray", "debug" },
          "Run stray light CKD generation in debug mode",
          SettingLevel::debug },
        { { "fov", "debug" },
          "Run field of view CKD generation in debug mode",
          SettingLevel::debug },
        { { "swath", "debug" },
          "Run swath vector CKD generation in debug mode",
          SettingLevel::debug },
        { { "wave", "debug" },
          "Run spectral CKD generation in debug mode",
          SettingLevel::debug },
        { { "rad", "debug" },
          "Run radiometric CKD generation in debug mode",
          SettingLevel::debug },
        { { "pol", "debug" },
          "Run polarimetric CKD generation in debug mode",
          SettingLevel::debug },
        { { "l1b", "debug" },
          "Run L1B process in debug mode",
          SettingLevel::debug },
        { { "l1c", "debug" },
          "Run L1C process in debug mode",
          SettingLevel::debug },
      } };
    std::array<Setting<int>, static_cast<int>(ProcLevel::n_levels)>
      van_cittert_steps { {
        {}, // l1a
        {}, // raw
        {}, // dark
        {}, // noise
        {}, // nonlin
        {}, // prnu
        {}, // stray
        {}, // fov
        {}, // swath
        {}, // wave
        { { "rad", "van_cittert_steps" }, 3, van_cittert_steps_text },
        { { "pol", "van_cittert_steps" }, 3, van_cittert_steps_text },
        { { "l1b", "van_cittert_steps" }, 3, van_cittert_steps_text },
        {}, // l1c
      } };
    std::array<Setting<std::string>, static_cast<int>(ProcLevel::n_levels)>
      ckd_file_in { {
        {}, // l1a
        {}, // raw
        { { "dark", "ckd_file_in" },
          {},
          ckd_file_in_text,
          SettingLevel::debug },
        { { "noise", "ckd_file_in" },
          {},
          ckd_file_in_text,
          SettingLevel::debug },
        { { "nonlin", "ckd_file_in" },
          {},
          ckd_file_in_text,
          SettingLevel::debug },
        { { "prnu", "ckd_file_in" },
          {},
          ckd_file_in_text,
          SettingLevel::debug },
        { { "stray", "ckd_file_in" },
          {},
          ckd_file_in_text,
          SettingLevel::debug },
        { { "fov", "ckd_file_in" }, {}, ckd_file_in_text, SettingLevel::debug },
        { { "swath", "ckd_file_in" },
          {},
          ckd_file_in_text,
          SettingLevel::debug },
        { { "wave", "ckd_file_in" },
          {},
          ckd_file_in_text,
          SettingLevel::debug },
        { { "rad", "ckd_file_in" }, {}, ckd_file_in_text, SettingLevel::debug },
        { { "pol", "ckd_file_in" }, {}, ckd_file_in_text, SettingLevel::debug },
        {}, // l1b
        {}, // l1c
      } };

    // Variables that are inferred from the input configuration files
    // but are not directly editable by user
    std::array<bool, static_cast<int>(ProcLevel::n_levels)> skip {};

    SPEXSettings() = default;
    SPEXSettings(const std::string& yaml_file) : Settings { yaml_file } {}
    auto scanKeys() -> void override;
    // Return a clone of the configuration with nodes removed that are
    // not present in [main][processes]. For instance, if swath is not
    // in [main][processes] then the swath section is removed even if
    // specified in the input configuration file.
    auto getReducedConfig() const -> std::string;
    ~SPEXSettings() = default;
};

} // namespace spex
