// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "theodolite.h"

#include "algorithm.h"
#include "linalg.h"
#include "nc_object.h"

#include <algorithm>
#include <gadfit/lm_solver.h>
#include <numbers>
#include <ranges>
#include <spdlog/spdlog.h>

namespace spex {

// Read theodolite file
static auto readTheodolite(const std::string filename,
                           std::vector<double>& act,
                           std::vector<double>& alt,
                           std::vector<int>& tar,
                           std::vector<double>& ang,
                           std::vector<double>& stst) -> void
{
    netCDF::NcFile nc { filename, netCDF::NcFile::read };
    const int nmeas { static_cast<int>(nc.getDim("nmeas").getSize()) };
    act.resize(nmeas);
    alt.resize(nmeas);
    std::vector<int> rail(nmeas * Theodolite::n_tar);
    // Targets of all measurements
    std::vector<int> tar_full(nmeas * Theodolite::n_tar);
    // All measurements must have consistent targets. In the end, one
    // set of targets is returned.
    tar.resize(Theodolite::n_tar);
    ang.resize(nmeas * Theodolite::n_tar * Theodolite::n_hv);
    nc.getVar("stage_act_angle").getVar(act.data());
    nc.getVar("stage_alt_angle").getVar(alt.data());
    nc.getVar("rail_number").getVar(rail.data());
    nc.getVar("target").getVar(tar_full.data());
    nc.getVar("theodolite_angles").getVar(ang.data());
    for (int i {}; i < nmeas; ++i) {
        act[i] *= math::deg_to_rad;
        alt[i] *= math::deg_to_rad;
    }
    for (double& val : ang) {
        val *= math::deg_to_rad;
    }
    // Put all measurements on rail 1 by shifting H for all
    // measurements on rail 2.
    for (int i {}; i < nmeas * Theodolite::n_tar; ++i) {
        if (rail[i] == 2) {
            ang[i * Theodolite::n_hv] += 90.0 * math::deg_to_rad;
        }
    }
    // Keep what we need. Throw away the rest.
    tar = { tar_full[0], tar_full[1] };
    // Optional: The star stimulus
    if (const auto nc_stst { nc.getVar("star_stimulus") }; nc_stst.isNull()) {
        return;
    } else {
        stst.resize(Theodolite::n_hv);
        nc_stst.getVar(stst.data());
    }
    // Also read the rail number. If there is a star stimulus, there
    // must also be rail associated to it.
    int stst_rail {};
    nc.getVar("star_stimulus_rail_number").getVar(&stst_rail);
    for (double& val : stst) {
        val *= math::deg_to_rad;
    }
    // Correct for rail number
    if (stst_rail == 2) {
        stst[0] += 90.0 * math::deg_to_rad;
    }
}

// Turn theodolite angles into orientation quaternion
static auto theoToQuat(const std::vector<int>& tar,
                       const std::vector<double>& theo_angs,
                       Quaternion& quat) -> void
{
    // Instrument alignment cube face normal vectors in the laboratory
    // refence frame
    std::array<std::array<double, dims::vec>, dims::vec> basis_n {};
    // Count the number of times a negative axis is encountered
    int n_sign_flips { 1 };
    // Record the axis in the order they are encountered
    std::array<int, dims::vec> axes {};
    for (int i_tar {}; i_tar < Theodolite::n_tar; ++i_tar) {
        // From the allowed values of target (-X, X, -Y, Y, -Z, Z)
        // extract the axis and sign.
        axes[i_tar] = { tar[i_tar] / 2 };
        const int sign { tar[i_tar] % 2 == 0 ? -1 : 1 };
        // X = cos(V)
        basis_n[axes[i_tar]][0] =
          sign * std::cos(theo_angs[i_tar * Theodolite::n_hv + 1]);
        // Y = cos(H) * sin(v)
        basis_n[axes[i_tar]][1] =
          sign * std::cos(theo_angs[i_tar * Theodolite::n_hv])
          * std::sin(theo_angs[i_tar * Theodolite::n_hv + 1]);
        // Z = sin(H) * sin(v)
        basis_n[axes[i_tar]][2] =
          sign * std::sin(theo_angs[i_tar * Theodolite::n_hv])
          * std::sin(theo_angs[i_tar * Theodolite::n_hv + 1]);
        n_sign_flips *= sign;
    }
    // Third axis is determined by the first two
    axes.back() = dims::vec - axes[0] - axes[1];
    // Third basis vector is a cross product of the first two
    crossProduct(basis_n[axes[0]], basis_n[axes[1]], basis_n[axes[2]]);
    // If one (but not two) of the vectors correpond to a negative
    // axis then also the cross product should be flipped.
    for (double& val : basis_n[axes[2]]) {
        val *= n_sign_flips;
    }
    // This does something only if there is a imperfection
    normalizeVector(basis_n[axes[2]]);
    // The two calculated base vectors may not be exactly
    // perpendicular. This is fixed by adapting the second target.
    crossProduct(basis_n[axes[2]], basis_n[axes[0]], basis_n[axes[1]]);
    for (double& val : basis_n[axes[1]]) {
        val *= n_sign_flips;
    }
    // A rotation matrix is formed from an orthonormal basis by
    // combining the three vectors and transposing.
    const std::array<double, dims::vec * dims::vec> rot {
        basis_n[0][0], basis_n[1][0], basis_n[2][0],
        basis_n[0][1], basis_n[1][1], basis_n[2][1],
        basis_n[0][2], basis_n[1][2], basis_n[2][2]
    };
    // Last step is to convert the base vectors into a Quaternion. It
    // may be a bit ambiguous in what direction we define this
    // quaternion. Are we rotating the axes or the vector inside the
    // axes, resulting in conjugate results. Check carefully if the
    // world is not turned upside down.
    quat = { rot };
}

// Get a unit vector from the matrix that belongs to a quaternion
static auto quatToUnitvec(const Quaternion& quat,
                          const int ax,
                          std::array<double, dims::vec>& vec) -> void
{
    //              n1               n2            n3
    // --------------------------------------------------------
    //     ( 1 - 2(j^2 + k^2)    2(ij-kr)        2(ik+jr)     )
    // R = (     2(ij + kr)  1 - 2(i^2 + k^2)    2(jk-ir)     )
    //     (     2(ik - jr)      2(jk+ir)    1 - 2(i^2 + j^2) )
    const double& i { quat[0] };
    const double& j { quat[1] };
    const double& k { quat[2] };
    const double& r { quat.real() };
    switch (ax) {
    case 0:
        vec[0] = 1.0 - 2.0 * (j * j + k * k);
        vec[1] = 2.0 * (i * j + k * r);
        vec[2] = 2.0 * (i * k - j * r);
        break;
    case 1:
        vec[0] = 2.0 * (i * j - k * r);
        vec[1] = 1.0 - 2.0 * (i * i + k * k);
        vec[2] = 2.0 * (j * k + i * r);
        break;
    case 2:
        vec[0] = 2.0 * (i * k + j * r);
        vec[1] = 2.0 * (j * k - i * r);
        vec[2] = 1.0 - 2.0 * (i * i + j * j);
        break;
    default:
        spdlog::warn("Undefined axis");
    }
}

// Do the exact opposite of calculateSwathVector: Turn orientation
// quaternion into theodolite angles. Input: orientation
// quaternion. Output: theodolite angles.
static auto quatToTheo(const std::vector<int>& tar,
                       const Quaternion& quat,
                       std::vector<double>& theo_angs) -> void
{
    for (int i_tar {}; i_tar < Theodolite::n_tar; ++i_tar) {
        const int ax { tar[i_tar] / 2 }; // 0=X, 1=Y, 2=Z
        std::array<double, dims::vec> vec {};
        quatToUnitvec(quat, ax, vec);
        // If the axis is in the negative direction flip the vector sign
        if (tar[i_tar] % 2 == 0) {
            for (double& val : vec) {
                val *= -1.0;
            }
        }
        // Convert them into theodolite angles
        // X = cos(V)
        // Y = cos(H) * sin(v)
        // Z = sin(H) * sin(v)
        theo_angs[i_tar * Theodolite::n_hv] = std::atan2(vec[2], vec[1]); // H
        theo_angs[i_tar * Theodolite::n_hv + 1] = std::acos(vec[0]);      // V
    }
}

auto Theodolite::stageToTheo(const double act,
                             const double alt,
                             std::vector<double>& theo_angs) const -> void
{
    // 1: Evaluate the ALT model to convert the ALT angle into an orientation
    //    quaternion for ACT = 0
    // 2: Convert this quaternion into H and V
    // 3: Use ACT slopes to include nonzero ACT influence
    // 4: Add offset in H and V from spot check file
    Quaternion quat { std::sin(alt / 2.0) * r_alt[0],
                      std::sin(alt / 2.0) * r_alt[1],
                      std::sin(alt / 2.0) * r_alt[2],
                      std::cos(alt / 2.0) };
    quat.multiply(quat0);
    quatToTheo(tar, quat, theo_angs);
    // Apply ACT derivative from target viewport and offset
    const auto target_vp { std::distance(
      alt_vp.begin(), std::ranges::find_if(alt_vp, [alt](const double x) {
          return std::abs(x - alt) < 10.0 * math::deg_to_rad;
      })) };
    for (int i {}; i < n_tar * n_hv; ++i) {
        theo_angs[i] += act * theo_act1[target_vp][i] + theo_offset[i];
    }
}

Theodolite::Theodolite(const std::vector<double>& viewport_alt_angles,
                       const std::string& reference_filename,
                       const std::string& spotcheck_filename)
{
    // PART 1 - Process the reference file
    const int n_vp { static_cast<int>(viewport_alt_angles.size()) };
    std::vector<double> act {};
    std::vector<double> alt {};
    std::vector<double> ang {};
    std::vector<double> stst {};
    readTheodolite(reference_filename, act, alt, tar, ang, stst);
    // Sort all measurements by ALT angle
    std::vector<std::vector<double>> act_vp(n_vp);
    alt_vp.resize(n_vp);
    std::vector<std::vector<double>> ang_vp(n_vp);
    for (int i {}; i < static_cast<int>(act.size()); ++i) {
        const double alt_value { alt[i] };
        const int i_vp { static_cast<int>(std::distance(
          viewport_alt_angles.cbegin(),
          std::ranges::find_if(viewport_alt_angles, [alt_value](const auto x) {
              constexpr double diff_thres { 1.0 * math::deg_to_rad };
              return std::abs(x - alt_value) < diff_thres;
          }))) };
        act_vp[i_vp].push_back(act[i]);
        alt_vp[i_vp] = alt[i];
        for (int k {}; k < n_tar * n_hv; ++k) {
            ang_vp[i_vp].push_back(ang[i * n_tar * n_hv + k]);
        }
    }
    // Fit linear relationship between theodolite and ACT stage angles
    // for each viewport.
    theo_act1.resize(n_vp, std::vector<double>(n_tar * n_hv));
    // Offsets for ALT model
    std::vector<std::vector<double>> theo_act0(
      n_vp, std::vector<double>(n_tar * n_hv));
    for (int i_vp {}; i_vp < n_vp; ++i_vp) {
        std::vector<double> y_data(act_vp[i_vp].size());
        // i_iac refers to an IAC face
        for (int i_iac {}; i_iac < n_tar * n_hv; ++i_iac) {
            for (int i {}; i < static_cast<int>(y_data.size()); ++i) {
                y_data[i] = ang_vp[i_vp][i * n_tar * n_hv + i_iac];
            }
            simpleLinearRegression(act_vp[i_vp],
                                   y_data,
                                   theo_act1[i_vp][i_iac],
                                   theo_act0[i_vp][i_iac]);
        }
    }
    // Next, solve the ALT model. The measurements are the IAC
    // orientations (expressed as matrix or quaternion) for each
    // viewport with ACT=0.  The state is a zero-orientation for ALT=0
    // and a rotation axis with which it is rotated about with
    // prescribed magnitude.
    // First, convert the five measurements into quaternions
    std::vector<Quaternion> alt_model_iacs(n_vp);
    for (int i_vp {}; i_vp < n_vp; ++i_vp) {
        theoToQuat(tar, theo_act0[i_vp], alt_model_iacs[i_vp]);
    }
    // The data for the nonlinear solver are quaternions - 4-component
    // objects - must be flattened first. As x-data, use an integer
    // array which is resolved into viewport number and quaternion
    // component number inside the solver. As y-data, use the
    // quaternion components of each viewport.
    std::vector<double> x_data(n_vp * dims::quat);
    std::iota(x_data.begin(), x_data.end(), 0.0);
    std::vector<double> y_data(n_vp * dims::quat);
    for (int i_vp {}; i_vp < n_vp; ++i_vp) {
        for (int i_dir {}; i_dir < dims::vec; ++i_dir) {
            y_data[i_vp * dims::quat + i_dir] = alt_model_iacs[i_vp][i_dir];
        }
        y_data[(i_vp + 1) * dims::quat - 1] = alt_model_iacs[i_vp].real();
    }
    gadfit::LMsolver solver { [&viewport_alt_angles](
                                const std::vector<gadfit::AdVar>& parameters,
                                const double x) {
        const int i_x { static_cast<int>(x) };
        const int i_vp { i_x / dims::quat };
        const int i_comp { i_x - i_vp * dims::quat };
        const double beta_2 { 0.5 * viewport_alt_angles[i_vp] };
        int idx { -1 };
        // Base quaternion (Q0) components
        const gadfit::AdVar& qx { parameters[++idx] };
        const gadfit::AdVar& qy { parameters[++idx] };
        const gadfit::AdVar& qr { parameters[++idx] };
        const gadfit::AdVar qz { sqrt(1.0 - qx * qx - qy * qy - qr * qr) };
        // Rotation vector (r_ALT) components
        const gadfit::AdVar& vy_0 { parameters[++idx] };
        const gadfit::AdVar& vz_0 { parameters[++idx] };
        const gadfit::AdVar vx_0 { sqrt(1.0 - vy_0 * vy_0 - vz_0 * vz_0) };
        // Rotation vector (r_ALT) components scaled by a scalar
        const gadfit::AdVar vx { std::sin(beta_2) * vx_0 };
        const gadfit::AdVar vy { std::sin(beta_2) * vy_0 };
        const gadfit::AdVar vz { std::sin(beta_2) * vz_0 };
        // Real part of rotation quaternion: QALT = r + r_ALT
        const double r { std::cos(beta_2) };
        // Quaternion multiplication Q1 * Q2 where Q1 = QALT, Q2 = Q0:
        // Q1 * Q2 = (Q1.r * Q2.r - Q1.v * Q2.v,
        //            Q1.r * Q2.v + Q2.r * Q1.v + Q1.v * x Q2.v)
        switch (i_comp) {
        case 0:
            // (Q1 * Q2).x
            return r * qx + qr * vx + vy * qz - vz * qy;
        case 1:
            // (Q1 * Q2).y
            return r * qy + qr * vy + vz * qx - vx * qz;
        case 2:
            // (Q1 * Q2).z
            return r * qz + qr * vz + vx * qy - vy * qx;
        case 3:
            // (Q1 * Q2).r
            return r * qr - (vx * qx + vy * qy + vz * qz);
        default:
            return gadfit::AdVar { gadfit::passive_idx };
        }
    } };
    solver.addDataset(x_data, y_data);
    solver.setPar(0, 0.1, true); // qx
    solver.setPar(1, 0.1, true); // qy
    solver.setPar(2, 0.1, true); // qr
    solver.setPar(3, 0.1, true); // vy
    solver.setPar(4, 0.1, true); // vz
    solver.settings.verbosity = gadfit::io::hide_all;
    solver.fit();
    // Construct the model base quaternion (Q0) from fitting results
    // q_x
    quat0[0] = solver.getParValue(0);
    // q_y
    quat0[1] = solver.getParValue(1);
    // q_r
    quat0.real() = solver.getParValue(2);
    // q_z = sqrt(1 - q_x^2 - q_y^2 - q_r^2)
    quat0[2] = std::sqrt(1.0 - quat0[0] * quat0[0] - quat0[1] * quat0[1]
                         - quat0.real() * quat0.real());
    // Also construct the model rotation vector (r_ALT)
    // q_y
    r_alt[1] = solver.getParValue(3);
    // q_z
    r_alt[2] = solver.getParValue(4);
    // q_x = sqrt(1 - q_y^2 - q_z^2)
    r_alt[0] = std::sqrt(1.0 - r_alt[1] * r_alt[1] - r_alt[2] * r_alt[2]);

    // PART 2 - Add information from spot check file
    std::vector<int> tar_spot {};
    readTheodolite(spotcheck_filename, act, alt, tar_spot, ang, stst);
    // Get theodolite angles from model with zero offset. Compare them
    // with the actual theodolite angles and the difference becomes
    // the offset.
    std::vector<double> expected_angs(n_tar * n_hv);
    stageToTheo(act.front(), alt.front(), expected_angs);
    for (int i_tar {}; i_tar < n_tar; ++i_tar) {
        for (int i_hv {}; i_hv < n_hv; ++i_hv) {
            double diff { ang[i_tar * n_hv + i_hv]
                          - expected_angs[i_tar * n_hv + i_hv] };
            // Wrap the horizontal (longitude-like) difference
            if (i_hv == 0) {
                if (diff > std::numbers::pi) {
                    diff -= 2 * std::numbers::pi;
                }
                if (diff < -std::numbers::pi) {
                    diff += 2 * std::numbers::pi;
                }
            }
            theo_offset[i_tar * n_hv + i_hv] = diff;
        }
    }
    // Star stimulus
    // X = cos(V)
    star_stimulus[0] = std::cos(stst[1]); // 1 stands for V
    // Y = cos(H) * sin(v)
    star_stimulus[1] = std::cos(stst[0]) * std::sin(stst[1]);
    // Z = sin(H) * sin(v)
    star_stimulus[2] = std::sin(stst[0]) * std::sin(stst[1]);
}

auto Theodolite::calculateSwathVector(
  const double act,
  const double alt,
  std::array<double, dims::vec>& swathvector) const -> void
{
    // 1. Convert the stage into expected theodolite angles
    // 2. Convert these theodolite angles into a quaternion
    // 3. Compare this quaternion (=matrix) with the star stimulus direction
    //    vector
    std::vector<double> theo_angs(n_tar * n_hv);
    stageToTheo(act, alt, theo_angs);
    Quaternion quat;
    theoToQuat(tar, theo_angs, quat);
    for (int ax {}; ax < dims::vec; ++ax) {
        std::array<double, dims::vec> vec {};
        quatToUnitvec(quat, ax, vec);
        swathvector[ax] = dotProduct(star_stimulus, vec);
    }
}

} // namespace spex
