// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

// Class for processing theodolite measurement data and constructing
// the ALT model from them. This class allows to correct for slightly
// different experimental conditions between different measurements of
// the stage angles.

#pragma once

#include "quaternion.h"

namespace spex {

class Theodolite
{
public:
    // Number of targets where target is one of the values: 0: -X, 1: +X,
    // 2: -Y, 3: +Y, 4: -Z, 5: +Z.
    static constexpr int n_tar { 2 };
    // Number of horizontal/vertical theodolite angles
    static constexpr int n_hv { 2 };

    // Constructor does two things: process data from the theodolite
    // reference file and construct the ALT model, and add information
    // from spot check file.
    Theodolite(const std::vector<double>& viewport_alt_angles,
               const std::string& reference_filename,
               const std::string& spotcheck_filename);

    // Get swath vector from stage angle using the theodolite measurements
    auto calculateSwathVector(const double act,
                              const double alt,
                              std::array<double, dims::vec>& swathvector) const
      -> void;

    ~Theodolite() = default;

private:
    // Nominal ALT angle per viewport
    std::vector<double> alt_vp {};
    // First and second target
    std::vector<int> tar {};
    // Derivatives of theodolite angles with respect to ACT stage
    // angle for each target and viewport
    std::vector<std::vector<double>> theo_act1 {};
    // Base quaternion for the ALT angle dependence model
    Quaternion quat0 {};
    // Rotation axis for ALT angle dependence model
    std::array<double, dims::vec> r_alt {};
    // Results from spotcheck calculation (first evalution of stageToTheo is
    // with zero offset)
    std::array<double, n_tar * n_hv> theo_offset { 0.0, 0.0, 0.0, 0.0 };
    std::array<double, dims::vec> star_stimulus {};
    // Acquire thedolite angles that correspond to a certain stage position
    auto stageToTheo(const double act,
                     const double alt,
                     std::vector<double>& theo_angs) const -> void;
};

} // namespace spex
