// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "wave_gen.h"

#include "l1a_metadata.h"

#include <gadfit/lm_solver.h>

namespace spex {

auto WaveGen::processInit() -> void
{
    for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
        ckd.wave.full[i_pol].resize(ckd.fov.n_act,
                                    std::vector<double>(ckd.n_detector_cols));
        ckd.wave.peak_data[i_pol].resize(ckd.fov.n_act);
        ckd.wave.peak_fits[i_pol].resize(
          ckd.fov.n_act, std::vector<double>(ckd.n_detector_cols));
        ckd.wave.spectral_data[i_pol].resize(ckd.fov.n_act);
    }
}

auto WaveGen::prepareBatches() -> void
{
    batchViewport();
}

static auto automaticPeakDetection(const std::vector<double>& signal,
                                   std::vector<double>& peak_positions,
                                   std::vector<double>& peak_signals) -> void
{
    // Peaks are detected by searching for values that are higher than
    // a certain percentage of the maximum signal.
    const double max_value { *std::ranges::max_element(signal) };
    for (int i {}; i < static_cast<int>(signal.size()); ++i) {
        // Threshold for detecting a peak as a portion of maximum signal
        constexpr double peak_thres { 0.05 };
        if (signal[i] > peak_thres * max_value) {
            peak_positions.push_back(static_cast<double>(i));
            peak_signals.push_back(signal[i]);
            // Once a peak has been detected, skip this many values
            // before searching for the next peak.
            constexpr int peak_island { 20 };
            i += peak_island;
            // Refine detected wavelength and signal values
            for (int j { i - peak_island + 1 }; j < i + peak_island; ++j) {
                if (signal[j] > peak_signals.back()) {
                    peak_positions.back() = j;
                    peak_signals.back() = signal[j];
                }
            }
        }
    }
}

auto WaveGen::processBatch(const int i_vp) -> void
{
    spdlog::info("Generating spectral CKD for viewport: {} deg",
                 viewport_idx_to_angle.at(
                   batches.at(i_vp).l1a_products.front().meta->viewport));
    const auto& l1a { batches.at(i_vp).l1a_products.front() };
    // Work arrays for storing the laser peak positions (column
    // indices) and the corresponding radiances.
    std::vector<double> peak_positions {};
    std::vector<double> peak_radiances {};
    // First, automatically detect the rough positions and radiances
    // of the reference spectrum peaks.
    automaticPeakDetection(
      l1a.meta->refspec_radiance, peak_positions, peak_radiances);
    // Find the corresponding peak wavelengths
    std::vector<double> peak_wavelengths(peak_positions.size());
    for (int i {}; i < static_cast<int>(peak_positions.size()); ++i) {
        peak_wavelengths[i] =
          l1a.meta->refspec_wavelength[static_cast<int>(peak_positions[i])];
    }
    int n_peaks { static_cast<int>(peak_wavelengths.size()) };
    if (i_vp == 0) {
        spdlog::debug("Detected {} peaks", n_peaks);
    }
    // Next, refine the peak wavelenghts by Gaussian fits of the
    // reference spectrum. Number of Gaussians in the fitting
    // functions equals the number of peaks.
    const auto gauss { [n_peaks](const std::vector<gadfit::AdVar>& pars,
                                 const double x) {
        gadfit::AdVar result { 0.0, gadfit::passive_idx };
        int idx { -1 };
        for (int i_peak {}; i_peak < n_peaks; ++i_peak) {
            const gadfit::AdVar& x0 { pars[++idx] };
            const gadfit::AdVar& a0 { pars[++idx] };
            const gadfit::AdVar& sigma { pars[++idx] };
            const gadfit::AdVar diff { x0 - x };
            result += a0 * exp(-diff * diff / sigma);
        }
        const gadfit::AdVar& bgr { pars[++idx] };
        return result + bgr;
    } };
    gadfit::LMsolver solver { gauss };
    solver.addDataset(l1a.meta->refspec_wavelength, l1a.meta->refspec_radiance);
    int idx { -1 };
    for (int i_peak {}; i_peak < n_peaks; ++i_peak) {
        // Parameter names are only shown in debug mode
        const std::string peak_str { std::to_string(i_peak) };
        solver.setPar(
          ++idx, peak_wavelengths[i_peak], true, "Peak " + peak_str + "    x0");
        solver.setPar(
          ++idx, peak_radiances[i_peak], true, "Peak " + peak_str + "    a0");
        solver.setPar(++idx, 5.0, true, "Peak " + peak_str + " sigma");
    }
    solver.setPar(++idx, 0.0, true, "Background");
    if (i_vp == 0
        && settings.debug.at(static_cast<int>(ProcLevel::wave)).value()) {
        solver.settings.verbosity = gadfit::io::final_only;
        spdlog::debug("Gaussian fitting results:");
    } else {
        solver.settings.verbosity = gadfit::io::hide_all;
    }
    solver.fit();
    for (int i_peak {}; i_peak < n_peaks; ++i_peak) {
        peak_wavelengths[i_peak] = solver.getParValue(i_peak * 3);
    }
    // Add offsets to all peaks if requested.
    for (int i {}; i < n_peaks; ++i) {
        peak_wavelengths[i] += settings.wave.peak_offset;
    }
    // If the reference wavelengths were specified by the user, then
    // use those instead and ignore the fit results.
    if (!settings.wave.reference_wavelengths.empty()) {
        peak_wavelengths = settings.wave.reference_wavelengths;
        n_peaks = static_cast<int>(peak_wavelengths.size());
    }
    if (i_vp == 0) {
        for (int i {}; i < n_peaks; ++i) {
            spdlog::debug("Registered peak at {:8.3f} nm", peak_wavelengths[i]);
        }
    }
    // Define an array or pixels 0...2047 (or lower if binned) that
    // serves as x-values for the solver.
    std::vector<double> pixels(ckd.n_detector_cols);
    std::iota(pixels.begin(), pixels.end(), 0.0);
    // Standard deviations for weighting the fitting procedure. These
    // are either 1 or a large value if dealing with a bad pixel.
    std::vector<double> std_devs(ckd.n_detector_cols);
#pragma omp parallel for firstprivate(pixels, std_devs) private(               \
  peak_positions, peak_radiances)
    for (int i_fov = 0; i_fov < ckd.fov.n_act_vp[i_vp]; ++i_fov) {
        // Global FOV index
        const int i_fov_g { ckd.fov.start_vp[i_vp] + i_fov };
        // Extract spectra from the detector image and process both
        // polarization states separately.
        Spectra spectra {};
        l1a.extractSpectra(ckd, i_fov_g, spectra);
        for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
            fillHoles(spectra.mask[i_pol], spectra.signal[i_pol]);
            for (int i {}; i < ckd.n_detector_cols; ++i) {
                std_devs[i] = spectra.mask[i_pol][i] ? 1e100 : 1.0;
            }
            // Reuse arrays for peak positions (columns) and radiances
            // but keep the peak wavelengths for later fitting.
            peak_positions.clear();
            peak_radiances.clear();
            automaticPeakDetection(
              spectra.signal[i_pol], peak_positions, peak_radiances);
            gadfit::LMsolver solver_gauss { gauss };
            solver_gauss.addDataset(pixels, spectra.signal[i_pol], std_devs);
            int idx { -1 };
            for (int i_peak {}; i_peak < n_peaks; ++i_peak) {
                solver_gauss.setPar(++idx, peak_positions[i_peak], true);
                solver_gauss.setPar(++idx, peak_radiances[i_peak], true);
                solver_gauss.setPar(++idx, 5.0, true);
            }
            solver_gauss.setPar(++idx, 0.0, true);
            solver_gauss.settings.verbosity = gadfit::io::hide_all;
            solver_gauss.fit();
            for (int i_peak {}; i_peak < n_peaks; ++i_peak) {
                peak_positions[i_peak] = solver_gauss.getParValue(i_peak * 3);
            }
            ckd.wave.peak_data[i_pol][i_fov_g] = spectra.signal[i_pol];
            for (int i {}; i < ckd.n_detector_cols; ++i) {
                ckd.wave.peak_fits[i_pol][i_fov_g][i] =
                  solver_gauss.getValue(static_cast<double>(i));
            }
            // Peak positions for the given spectrum have been
            // found. Next, fit the relation position:wavelength and
            // assign wavelengths to all columns.
            const int order { settings.wave.order };
            // Polynomial fitting function
            const auto poly { [order](const std::vector<gadfit::AdVar>& pars,
                                      const double x_in) {
                std::vector<double> x(order + 1);
                x.front() = 1.0;
                for (int i { 1 }; i <= order; ++i) {
                    x[i] = x[i - 1] * x_in;
                }
                gadfit::AdVar result { 0.0, gadfit::passive_idx };
                for (int i {}; i <= order; ++i) {
                    result += pars[i] * x[i];
                }
                return result;
            } };
            gadfit::LMsolver solver_poly { poly };
            solver_poly.addDataset(peak_positions, peak_wavelengths);
            for (int i {}; i <= order; ++i) {
                solver_poly.setPar(i, 0.0, true);
            }
            solver_poly.settings.verbosity = gadfit::io::hide_all;
            solver_poly.fit();
            for (int i {}; i < ckd.n_detector_cols; ++i) {
                ckd.wave.full[i_pol][i_fov_g][i] =
                  solver_poly.getValue(static_cast<double>(i));
            }
            // The fit data is stored in the CKD in debug mode
            ckd.wave.spectral_data[i_pol][i_fov_g] = peak_positions;
            if (i_fov == 0 && i_pol == 0) {
                ckd.wave.spectral_wavelengths = peak_wavelengths;
            }
        }
    }
}

auto WaveGen::processFinalize() -> void
{
    ckd.wave.common.resize(ckd.fov.n_act,
                           std::vector<double>(ckd.n_detector_cols));
    for (int i_fov {}; i_fov < ckd.fov.n_act; ++i_fov) {
        for (int i {}; i < ckd.n_detector_cols; ++i) {
            ckd.wave.common[i_fov][i] =
              0.5
              * (ckd.wave.full[pol_plus][i_fov][i]
                 + ckd.wave.full[pol_minus][i_fov][i]);
        }
    }
}

} // namespace spex
